@@ -172,10 +172,13 @@
 //! ```
 
 use std::convert::{TryFrom, TryInto};
+use std::ops::Range;
 
+use crate::de_arrays::ArrayWrapperAccess;
 use crate::error::{Error, Result};
 use crate::Tag;
 use byteorder::{BigEndian, ReadBytesExt};
+
 use serde::{de, forward_to_deserialize_any};
 
 /// Deserialize into a `T` from some NBT data. See the [`de`] module for more
@@ -205,7 +208,6 @@ where
 {
     let mut des = Deserializer::from_bytes(&input);
     let t = T::deserialize(&mut des)?;
-    // TODO: trailing chars?
     Ok(t)
 }
 
@@ -213,8 +215,9 @@ where
 ///
 /// [`de`]: ./index.html
 pub struct Deserializer<'de> {
-    input: InputHelper<'de>,
+    pub(crate) input: InputHelper<'de>,
     layers: Vec<Layer>,
+    last_hint: Option<&'static str>,
 }
 
 impl<'de> Deserializer<'de> {
@@ -226,6 +229,7 @@ impl<'de> Deserializer<'de> {
         Self {
             input: InputHelper(input),
             layers: vec![],
+            last_hint: None,
         }
     }
 }
@@ -251,12 +255,15 @@ enum Layer {
 /// input. If we wrote the helper functions as part of the Deserializer impl, it
 /// would force borrowing the entire deserializer mutably. This helper allows us
 /// to borrow just the input, making us free to also borrow/mutate the layers.
-struct InputHelper<'de>(&'de [u8]);
+pub(crate) struct InputHelper<'de>(pub(crate) &'de [u8]);
 
 fn consume_value<'de, V>(de: &mut Deserializer<'de>, visitor: V, tag: Tag) -> Result<V::Value>
 where
     V: de::Visitor<'de>,
 {
+    let last_hint = de.last_hint;
+    de.last_hint = None;
+
     match tag {
         Tag::Byte => visitor.visit_i8(de.input.0.read_i8()?),
         Tag::Short => visitor.visit_i16(de.input.0.read_i16::<BigEndian>()?),
@@ -277,6 +284,20 @@ where
             let element_tag = de.input.consume_tag()?;
             let size = de.input.consume_list_size()?;
 
+            // End values have no payload. An end tag on it's own is the payload
+            // of an empty compound. A logical interpretation is that this could
+            // be a list of zero-sized units, but this mean an easy short
+            // malicious payload of a massive list taking up lots of memory (as
+            // the Value type's unit variant would not be zero sized.
+            //
+            // Some old chunks store empty lists as as 'list of end', so if the
+            // size is zero we let it slide.
+            if element_tag == Tag::End && size != 0 {
+                return Err(Error::bespoke(
+                    "unexpected list of type 'end', which is not supported".into(),
+                ));
+            }
+
             de.layers.push(Layer::List {
                 remaining_elements: size,
                 element_tag,
@@ -285,27 +306,34 @@ where
             visitor.visit_seq(ListAccess::new(de, size))
         }
         Tag::ByteArray | Tag::IntArray | Tag::LongArray => {
+            if last_hint == Some("seq") {
+                return Err(Error::bespoke(
+                    "expected NBT Array, found seq: use ByteArray, IntArray or LongArray types"
+                        .into(),
+                ));
+            }
             let size = de.input.consume_list_size()?;
-            let non_array_tag = match tag {
-                Tag::ByteArray => Tag::Byte,
-                Tag::IntArray => Tag::Int,
-                Tag::LongArray => Tag::Long,
-                _ => panic!(),
-            };
-
-            de.layers.push(Layer::List {
-                remaining_elements: size,
-                element_tag: non_array_tag,
-            });
-
-            // Going to pretend we're in a list to reuse the ListAccess.
-            visitor.visit_seq(ListAccess::new(de, size))
+            visitor.visit_map(ArrayWrapperAccess::new(de, size, tag))
         }
-        _ => todo!("any value"),
+        // This would really only occur when we encounter a list where the
+        // element type is 'End', but we specifically handle that case, so we
+        // should never get here.
+        Tag::End => Err(Error::bespoke(
+            "unexpected end tag, was expecting payload of a value".into(),
+        )),
     }
 }
 
 impl<'de> InputHelper<'de> {
+    // Safely get a subslice, erroring if there's not enough input.
+    pub(crate) fn subslice(&self, r: Range<usize>) -> Result<&'de [u8]> {
+        if r.end <= self.0.len() {
+            Ok(&self.0[r])
+        } else {
+            Err(Error::unexpected_eof())
+        }
+    }
+
     fn consume_tag(&mut self) -> Result<Tag> {
         let tag_byte = self.0.read_u8()?;
         Tag::try_from(tag_byte).or_else(|_| Err(Error::invalid_tag(tag_byte)))
@@ -317,7 +345,7 @@ impl<'de> InputHelper<'de> {
 
     fn consume_size_prefixed_string(&mut self) -> Result<&'de str> {
         let len = self.0.read_u16::<BigEndian>()? as usize;
-        let s = std::str::from_utf8(&self.0[..len])
+        let s = std::str::from_utf8(self.subslice(0..len)?)
             .map_err(|_| Error::nonunicode_string(&self.0[..len]));
         self.0 = &self.0[len..];
         s
@@ -417,7 +445,7 @@ impl<'de> InputHelper<'de> {
 impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
     type Error = Error;
 
-    forward_to_deserialize_any!(struct map identifier i8 i16 i32 i64 u8 u16 u32 u64 f32 f64 str string seq tuple);
+    forward_to_deserialize_any!(struct map identifier i8 i16 i32 i64 u8 u16 u32 u64 f32 f64 str string tuple);
 
     fn is_human_readable(&self) -> bool {
         false
@@ -708,6 +736,16 @@ impl<'de, 'a> de::Deserializer<'de> for &'a mut Deserializer<'de> {
 
         visitor.visit_unit()
     }
+
+    fn deserialize_seq<V>(self, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        // We record the fact we saw seq, so that if we then try to decode an
+        // NBT array, we can provide a nice error message.
+        self.last_hint = Some("seq");
+        self.deserialize_any(visitor)
+    }
 }
 
 struct CompoundAccess<'a, 'de> {
