@@ -4,8 +4,11 @@ use crate::Tag;
 
 mod builder;
 mod de;
+mod fuzz;
 mod minecraft_chunk;
 mod stream;
+mod de_arrays;
+mod value;
 
 fn assert_try_into(tag: Tag) {
     assert_eq!(tag, (tag as u8).try_into().unwrap());
